// This file must be the first include for any bindings code.

#pragma once

#include <nanobind/nanobind.h>

namespace mqt {
namespace nb = nanobind;
using namespace nb::literals;

} // namespace mqt
